#  Copyright 2024 Pressinfra SRL
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Optional

import uuid6

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.db import models, transaction
from django.db.models import Q, UniqueConstraint

from django_woah.utils.models import AutoCleanModel


class UserGroupKind(models.TextChoices):
    ROOT = "root", "Root"
    TEAM = "team", "Team"
    USER = "user", "User"


class UserGroup(AutoCleanModel):
    KINDS = UserGroupKind

    id = models.UUIDField(default=uuid6.uuid7, unique=True, primary_key=True)

    name = models.CharField(max_length=80, blank=True)
    kind = models.CharField(choices=UserGroupKind.choices, max_length=16)

    # TODO: should this be nullable?
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="owned_user_groups",
        on_delete=models.CASCADE,
    )

    parent = models.ForeignKey(
        "self", related_name="children", null=True, blank=True, on_delete=models.CASCADE
    )
    # This is some yet to be decided denormalization
    root = models.ForeignKey(
        "self",
        related_name="root_descendants",
        limit_choices_to=Q(kind=UserGroupKind.ROOT),
        null=True,
        blank=True,
        on_delete=models.CASCADE,
    )

    parent_membership = models.OneToOneField(
        "django_woah.Membership", null=True, blank=True, on_delete=models.CASCADE
    )

    # This is some denormalization, but it's possibly useful
    related_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="related_user_groups",
        null=True,
        blank=True,
        on_delete=models.CASCADE,
    )

    class Meta:
        constraints = [
            UniqueConstraint(
                fields=("owner", "kind", "related_user"),
                name="unique_usergroup_owner_kind_related_user",
            ),
            UniqueConstraint(
                fields=("owner", "parent", "kind", "name"),
                name="unique_usergroup_localized_name",
            ),
        ]

    def clean(self):
        if self.root and not self.parent:
            raise ValidationError(
                "A parent must be specified for a non-root UserGroup."
            )

        if self.kind == UserGroupKind.USER and not self.parent_membership:
            raise ValidationError(
                "A parent membership must be specified for a user UserGroup."
            )

        if not self.name:
            username = ""
            if self.kind == UserGroupKind.USER:
                username = f"user:{str(self.related_user)} "

            self.name = " ".join(
                element for element in [str(self.owner), username, self.kind] if element
            )

    def __str__(self):
        name = f"{self.name}"

        return name


class AssignedPermManager(models.Manager):
    use_for_related_fields = True

    def get_queryset(self):
        return super().get_queryset().select_related("owner")

    def filter(self, *args, **kwargs):
        if kwargs and (resource := kwargs.pop("resource", None)):
            kwargs["content_type"] = ContentType.objects.get_for_model(
                resource.__class__
            )
            kwargs["object_id"] = resource.pk

        return super().filter(*args, **kwargs)

    def get(self, *args, **kwargs):
        if kwargs and (resource := kwargs.pop("resource", None)):
            kwargs["content_type"] = ContentType.objects.get_for_model(
                resource.__class__
            )
            kwargs["object_id"] = resource.pk

        return super().get(*args, **kwargs)

    def create(self, defaults=None, **kwargs):
        if resource := kwargs.pop("resource", None):
            kwargs["content_type"] = ContentType.objects.get_for_model(
                resource.__class__
            )
            kwargs["object_id"] = resource.pk

        return super().create(**kwargs)

    def get_or_create(self, defaults=None, **kwargs):
        if defaults and (resource := defaults.pop("resource", None)):
            defaults["content_type"] = ContentType.objects.get_for_model(
                resource.__class__
            )
            defaults["object_id"] = resource.pk

        if resource := kwargs.pop("resource", None):
            kwargs["content_type"] = ContentType.objects.get_for_model(
                resource.__class__
            )
            kwargs["object_id"] = resource.pk

        return super().get_or_create(defaults=defaults, **kwargs)


# class AssignedPermAppliesTo(models.TextChoices):
#     ALL = "all", "All"
#     MEMBERS = "members", "Members"
#     OUTSIDE_COLLABORATORS = "outside_collaborators", "Outside Collaborators"


class AssignedPerm(AutoCleanModel):
    id = models.UUIDField(default=uuid6.uuid7, unique=True, primary_key=True)

    user_group = models.ForeignKey(
        UserGroup, on_delete=models.CASCADE, related_name="group_assigned_perms"
    )
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="owned_assigned_perms",
        on_delete=models.CASCADE,
    )
    perm = models.CharField(max_length=128)

    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, null=True, blank=True
    )
    object_id = models.CharField(null=True, blank=True, max_length=40)
    resource = GenericForeignKey("content_type", "object_id")

    non_model_resource_id = models.TextField(null=True, blank=True)

    # restriction = models.TextField(null=True, blank=True)

    objects = AssignedPermManager()

    class Meta:
        indexes = [
            models.Index(fields=["content_type", "object_id"]),
        ]

    def __str__(self):
        target = "*" if not self.object_id else f"{self.content_type, self.object_id}"

        return f"<{self.perm}>, target:<{target}>, for:<{self.user_group}>"

    def full_clean(self, *args, **kwargs):
        resource_data = [self.content_type, self.object_id]
        if any(resource_data) and not all(resource_data):
            raise ValidationError(
                "Both or neither content_type and self.object_id must be specified."
            )

        if self.resource and self.non_model_resource_id:
            raise ValidationError(
                "Both a resource and a non_model_resource_id may not be specified."
            )

        try:
            self.owner
        except AssignedPerm.owner.RelatedObjectDoesNotExist:
            self.owner = self.user_group.owner

        return super().full_clean(*args, **kwargs)


class Membership(AutoCleanModel):
    id = models.UUIDField(default=uuid6.uuid7, unique=True, primary_key=True)

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name="memberships"
    )

    user_group = models.ForeignKey(
        UserGroup, on_delete=models.CASCADE, related_name="memberships"
    )

    # This is some yet to be decided denormalization
    root_user_group = models.ForeignKey(
        UserGroup,
        on_delete=models.CASCADE,
        related_name="root_memberships",
        limit_choices_to=Q(kind=UserGroupKind.ROOT),
    )

    is_outside_collaborator = models.BooleanField(
        default=False,
        help_text="Outside collaborators might receive different permissions than members, or even no permissions.",
    )

    parent = models.ForeignKey("self", null=True, blank=True, on_delete=models.CASCADE)

    class Meta:
        constraints = [
            UniqueConstraint(
                fields=("user", "user_group"), name="unique_user_to_group_membership"
            ),
        ]

    def clean(self):
        if self.user_group.parent and not self.parent:
            raise ValidationError(
                "A parent membership must be specified for a membership to a non-Root UserGroup."
            )

        if self.user_group.kind == UserGroupKind.USER:
            self.is_outside_collaborator = False

            if not self.user == self.user_group.related_user:
                raise ValidationError(
                    "User must match with the UserGroup(kind=USER) related_user."
                )

        root_user_group = self.user_group.root or self.user_group
        if not root_user_group:
            raise ValidationError("A root_user_group must be specified.")

        self.root_user_group = root_user_group


def get_root_user_group(owner) -> UserGroup:
    return UserGroup.objects.get(
        kind=UserGroupKind.ROOT,
        owner=owner,
        related_user=owner,
    )


def get_single_user_user_group(related_to_user, owned_by_account) -> UserGroup:
    return related_to_user.related_user_groups.get(owner=owned_by_account)


def get_or_create_root_user_group(owner) -> UserGroup:
    return UserGroup.objects.get_or_create(
        kind=UserGroupKind.ROOT,
        owner=owner,
        related_user=owner,
    )[0]


def get_or_create_team_user_group(owner, name: str) -> UserGroup:
    root = get_root_user_group(owner)

    return UserGroup.objects.get_or_create(
        name=name,
        kind=UserGroupKind.TEAM,
        owner=owner,
        root=root,
        parent=root,
    )[0]


def get_root_membership(user, account) -> Optional[Membership]:
    try:
        return Membership.objects.get(
            user=user,
            user_group__owner=account,
            user_group__kind=UserGroupKind.ROOT,
        )
    except Membership.DoesNotExist:
        return None


@transaction.atomic
def add_user_to_user_group(
    user, user_group: UserGroup, is_outside_collaborator=False
) -> tuple[Membership, UserGroup]:
    if not user_group.root and user_group.kind != UserGroupKind.ROOT:
        raise ValueError("Received a non-Root UserGroup with missing `root` field")

    parent_membership = None
    if user_group.parent:
        parent_membership = user_group.parent.memberships.get(user=user)

    root_user_group = user_group.root or user_group

    resulted_membership = Membership.objects.update_or_create(
        user=user,
        user_group=user_group,
        root_user_group=root_user_group,
        parent=parent_membership,
        defaults={"is_outside_collaborator": is_outside_collaborator},
    )[0]

    resulted_user_group = None
    if user_group.kind == UserGroupKind.ROOT:
        resulted_user_group = UserGroup.objects.update_or_create(
            kind=UserGroupKind.USER,
            parent=user_group,
            root=user_group,
            owner=user_group.owner,
            related_user=user,
            parent_membership=resulted_membership,
        )[0]

        Membership.objects.update_or_create(
            user=user,
            user_group=resulted_user_group,
            root_user_group=root_user_group,
            is_outside_collaborator=False,
            parent=resulted_membership,
        )

    return resulted_membership, resulted_user_group


def create_user_membership_to_account(
    user, account, is_outside_collaborator=False
) -> tuple[Membership, UserGroup]:
    user_group = get_root_user_group(account)

    return add_user_to_user_group(user, user_group, is_outside_collaborator)


def assign_perm(perm, to_user, on_account, for_resource=None) -> AssignedPerm:
    try:
        user_group = get_single_user_user_group(
            related_to_user=to_user, owned_by_account=on_account
        )
    except ObjectDoesNotExist:
        error_message = f"User {to_user} does not have a membership on {on_account}."
        solution = f"To grant the user membership, you may call `create_user_membership_to_account({to_user}, {on_account})`."

        print(f"{error_message}\n{solution}")
        raise ValueError(error_message)

    kwargs = {
        "user_group": user_group,
        "perm": perm,
    }
    if for_resource:
        kwargs["resource"] = for_resource

    return AssignedPerm.objects.create(
        user_group=get_single_user_user_group(
            related_to_user=to_user, owned_by_account=on_account
        ),
        perm=perm,
    )



#  Copyright 2024 Pressinfra SRL
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from django.db.models import Q, Model
from typing import Optional

from django_woah.utils.q import merge_qs, prefix_q_with_relation, get_object_relation
from .conditions import Condition
from .context import Context
from .enum import PermEnum


class IndirectPerms:
    def __init__(self, **kwargs):
        self.scheme = kwargs.get("scheme")

    def can_receive_perms(self) -> list[PermEnum]:
        """This function is mainly used to avoid unnecessary queries and some infinite recursions."""
        return []

    def get_resources_q(self, context: Context) -> Optional[Q]:
        raise NotImplementedError

    def get_assigned_perms_q(self, context: Context) -> Optional[Q]:
        raise NotImplementedError

    def get_memberships_q(self, context: Context) -> Optional[Q]:
        raise NotImplementedError

    def is_authorized_for_prefetched_resource(self, context: Context) -> bool:
        raise NotImplementedError

    def is_authorized_for_unsaved_resource(self, context: Context) -> bool:
        raise NotImplementedError

    def set_scheme(self, scheme):
        if self.scheme is None:
            self.scheme = scheme


class ConditionalPerms(IndirectPerms):
    def __init__(
        self, conditions: list[Condition], receives_perms: list[PermEnum], **kwargs
    ):
        self.conditions: list[Condition] = conditions
        self.receives_perms: list[PermEnum] = receives_perms

        super().__init__(**kwargs)

    def set_scheme(self, scheme):
        super().set_scheme(scheme)

        for condition in self.conditions:
            condition.set_scheme(self.scheme)

    def get_resources_q(self, context: Context) -> Optional[Q]:
        return merge_qs(
            [condition.get_resources_q(context) for condition in self.conditions]
        )

    def get_assigned_perms_q(self, context: Context) -> Optional[Q]:
        return merge_qs(
            [condition.get_assigned_perms_q(context) for condition in self.conditions],
            connector=Q.OR,
        )

    def get_memberships_q(self, context: Context) -> Optional[Q]:
        return merge_qs(
            [condition.get_memberships_q(context) for condition in self.conditions],
            connector=Q.OR,
        )

    def can_receive_perms(self):
        return self.receives_perms

    def is_authorized_for_prefetched_resource(self, context: Context) -> bool:
        return all(
            condition.is_authorized_for_prefetched_resource(context)
            for condition in self.conditions
        )

    def is_authorized_for_unsaved_resource(self, context: Context) -> bool:
        return all(
            condition.is_authorized_for_unsaved_resource(context)
            for condition in self.conditions
        )

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.conditions} -> {self.receives_perms}"


class TransitiveFromRelationPerms(IndirectPerms):
    def __init__(self, relation: str, restrict_to_perms=None, **kwargs):
        super().__init__(**kwargs)

        self.relation = relation
        self.restrict_to_perms = set(restrict_to_perms or [])

    def set_scheme(self, scheme):
        super().set_scheme(scheme)

        self.restrict_to_perms = self.restrict_to_perms.intersection(
            self.scheme.Perms.values()
        )
        self.relation_scheme = self.scheme.get_auth_scheme_by_relation(self.relation)

        if not self.restrict_to_perms:
            self.restrict_to_perms = set(self.scheme.Perms.values())

        relation_scheme_perms = set(self.relation_scheme.get_scheme_perms())

        if not self.restrict_to_perms.issubset(relation_scheme_perms):
            raise ValueError(
                f"{self.relation_scheme} has not borrowed the following perms: {self.restrict_to_perms.difference(relation_scheme_perms)}"
            )

    def get_resources_q(self, context: Context) -> Optional[Q]:
        if context.perm not in self.restrict_to_perms:
            raise ValueError(
                f"{context.perm} not in restrict_to_perms {self.restrict_to_perms}"
            )

        context = context.subcontext(resource=self.relation_scheme.model)

        q = self.relation_scheme.get_resources_q(context)
        if q is None:
            return None

        return prefix_q_with_relation(q, self.relation)

    def get_assigned_perms_q(self, context: Context) -> Optional[Q]:
        relation_resource = self.relation_scheme.model

        if isinstance(context.resource, self.scheme.model):
            concrete_relation_resource = get_object_relation(
                context.resource, self.relation
            )
            if concrete_relation_resource is None:
                return None

            if isinstance(concrete_relation_resource, Model):
                relation_resource = concrete_relation_resource

        return self.relation_scheme.get_assigned_perms_q(
            context=context.subcontext(resource=relation_resource)
        )

    def get_memberships_q(self, context: Context) -> Optional[Q]:
        relation_resource = self.relation_scheme.model

        if isinstance(context.resource, self.scheme.model):
            concrete_relation_resource = get_object_relation(
                context.resource, self.relation
            )
            if concrete_relation_resource is None:
                return None

            if isinstance(concrete_relation_resource, Model):
                relation_resource = concrete_relation_resource

        return self.relation_scheme.get_memberships_q(
            context=context.subcontext(resource=relation_resource)
        )

    def can_receive_perms(self) -> list[PermEnum]:
        return self.restrict_to_perms

    def is_authorized_for_prefetched_resource(self, context: Context) -> bool:
        if context.perm not in self.restrict_to_perms:
            raise ValueError(
                f"{context.perm} not in restrict_to_perms {self.restrict_to_perms}"
            )

        context = context.subcontext(
            resource=get_object_relation(context.resource, self.relation)
        )

        return self.scheme.auth_solver.is_authorized_for_prefetched_resource(context)

    def is_authorized_for_unsaved_resource(self, context: Context) -> bool:
        if context.perm not in self.restrict_to_perms:
            raise ValueError(
                f"{context.perm} not in restrict_to_perms {self.restrict_to_perms}"
            )

        context = context.subcontext(
            resource=get_object_relation(context.resource, self.relation)
        )

        solver = self.scheme.auth_solver
        context.assigned_perms = solver.get_assigned_perms_queryset(context)

        return solver.get_authorized_on_resources_queryset(context).exists()



#  Copyright 2024 Pressinfra SRL
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from functools import reduce
from typing import Optional
from typing import TYPE_CHECKING

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q, Model

from django_woah.models import AssignedPerm
from django_woah.utils.q import merge_qs, prefix_q_with_relation, pop_parts_of_q

if TYPE_CHECKING:
    from .solver import AuthorizationSolver

from .conditions import Condition
from .context import Context
from .enum import PermEnum
from .indirect_perms import IndirectPerms


class AuthorizationScheme:
    pass


class ModelAuthorizationScheme(AuthorizationScheme):
    owner_relation: str
    model: type[Model]
    Perms: PermEnum
    Roles: PermEnum
    allow_directly_assigned_perms = True

    def __init__(self):
        auth_solver: "AuthorizationSolver"  # noqa: F842

    def __init_subclass__(cls, **kwargs):
        for value in getattr(cls, "Perms", []):
            value.auth_scheme = cls
            value.is_permission = True
            value.is_role = False

        for value in getattr(cls, "Roles", None) or []:
            value.auth_scheme = cls
            value.is_permission = False
            value.is_role = True

    @classmethod
    def get_borrowed_perms(cls) -> list[PermEnum]:
        """
        Used to specify what perms of other schemes can be applied to this scheme.
        Necessary for when using TransitiveFromRelationPerms for example.
        """
        return []

    @classmethod
    def get_scheme_perms(cls, exclude_borrowed=False) -> list[PermEnum]:
        perms_and_roles = [] if exclude_borrowed else cls.get_borrowed_perms()

        if roles := getattr(cls, "Roles", []):
            perms_and_roles = list(roles) + perms_and_roles

        if perms := getattr(cls, "Perms", []):
            perms_and_roles = list(perms) + perms_and_roles

        return perms_and_roles

    def get_indirect_perms(self, context: Context) -> list[IndirectPerms]:
        return []

    def get_scheme_indirect_perms(self, context: Context) -> list[IndirectPerms]:
        indirect_perms = []
        for indirect_perm in self.get_indirect_perms(context):
            indirect_perm.set_scheme(self)
            indirect_perms.append(indirect_perm)

        return indirect_perms

    def get_implicit_conditions(self, context: Context) -> list[Condition]:
        return []

    def get_scheme_implicit_conditions(self, context: Context) -> list[Condition]:
        implicit_conditions = []
        for implicit_condition in self.get_implicit_conditions(context):
            implicit_condition.set_scheme(self)
            implicit_conditions.append(implicit_condition)

        return implicit_conditions

    def get_resources_q_from_directly_assigned_perms(
        self, context: Context
    ) -> Optional[Q]:
        # TODO:
        #  For when specific instances of models are queried, relying on assigned_perms
        #  to tell the truth here might cascade upstream in unexpected ways.
        #  So maybe using Context here as well and considering context.resource.pks, or somehow
        #  restricting to certain PKs could be the solution.

        if not self.allow_directly_assigned_perms:
            return None

        owner_based_q = None

        if hasattr(self, "owner_relation"):
            relation_model = self.get_model_for_relation(self.owner_relation)

            owner_based_matches = [
                assigned_perm.owner_id  # if relation_model != UserGroup else a.root_id
                for assigned_perm in context.assigned_perms
                if (
                    assigned_perm.perm == context.perm
                    and not assigned_perm.object_id
                    and not assigned_perm.content_type
                )
            ]

            if (pk_field_name := relation_model._meta.pk.name) == "id":
                owner_based_relation = (
                    "id" if self.owner_relation == "*" else f"{self.owner_relation}_id"
                )
            else:
                owner_based_relation = (
                    pk_field_name
                    if self.owner_relation == "*"
                    else f"{self.owner_relation}__{pk_field_name}"
                )

            owner_based_q = (
                Q(**{f"{owner_based_relation}__in": owner_based_matches})
                if owner_based_matches
                else None
            )

        directly_authorized_ids = [
            assigned_perm.object_id
            for assigned_perm in context.assigned_perms
            if (
                assigned_perm.perm == context.perm
                and assigned_perm.content_type == self.model_content_type
            )
        ]
        q = Q(pk__in=directly_authorized_ids) if directly_authorized_ids else None

        if q and owner_based_q:
            return q | owner_based_q

        return q or owner_based_q

    def get_resources_q_from_indirect_perms(self, context: Context) -> Optional[Q]:
        q = None

        for indirect_perms in self.get_scheme_indirect_perms(context):
            if context.perm in indirect_perms.can_receive_perms():
                if (
                    indirect_perms_q := indirect_perms.get_resources_q(context)
                ) is not None:
                    if not q:
                        q = indirect_perms_q
                    else:
                        q |= indirect_perms_q

        return q

    def get_resources_implicit_conditions_q(self, context: Context) -> Optional[Q]:
        """
        The resulting Q is meant to be AND-ed with other Qs, and not to be used on it's own.
        A `None` result means the Q is guaranteed to be logically False.
        """

        implicit_conditions_qs = [
            condition.get_resources_q(context)
            for condition in self.get_scheme_implicit_conditions(context) or []
        ]
        if not implicit_conditions_qs:
            return Q()

        return merge_qs(implicit_conditions_qs)

    def get_resources_q(self, context: Context) -> Optional[Q]:
        if context.assigned_perms is None:
            context.assigned_perms = AssignedPerm.objects.filter(
                self.get_assigned_perms_q(context)
            )

        q = merge_qs(
            [
                self.get_resources_q_from_directly_assigned_perms(context),
                self.get_resources_q_from_indirect_perms(context),
            ],
            connector=Q.OR,
        )

        q = merge_qs([q, self.get_resources_implicit_conditions_q(context)])

        # TODO check if this restriction is ok; also handle list of specific resources
        if isinstance(context.resource, Model):
            if not context.resource.pk:
                raise ValueError(
                    "Encountered context.resource without pk", context.resource
                )

            if q is not None:
                q &= Q(pk=context.resource.pk)

        return q

    def is_directly_authorized_for_prefetched_resource(self, context: Context) -> bool:
        # It is assumed that context.assigned_perms have been prefetched with scheme.get_assigned_perms_q()
        # or such that the User has membership to the UserGroups and what other implicit conditions the scheme
        # may specify.

        for assigned_perm in context.assigned_perms:
            if (
                assigned_perm.perm == context.perm
                # TODO: should allow assigned_perm.content_type=None too?
                and assigned_perm.content_type.model_class == context.resource.__class__
                and (
                    assigned_perm.object_id is None
                    or assigned_perm.object_id == context.resource.pk
                )
            ):
                return True

        return False

    def is_authorized_for_prefetched_resource(self, context: Context) -> bool:
        if context.assigned_perms is None:
            context.assigned_perms = AssignedPerm.objects.filter(
                self.get_assigned_perms_q(context)
            )

        for condition in self.get_scheme_implicit_conditions(context) or []:
            if not condition.is_authorized_for_prefetched_resource(context):
                return False

        if self.is_directly_authorized_for_prefetched_resource(context):
            return True

        for indirect_perm in self.get_scheme_indirect_perms(context):
            if context.perm not in indirect_perm.can_receive_perms():
                continue

            if indirect_perm.is_authorized_for_prefetched_resource(context):
                return True

        return False

    def is_authorized_for_unsaved_resource(self, context: Context) -> bool:
        for condition in self.get_scheme_implicit_conditions(context) or []:
            if not condition.is_authorized_for_unsaved_resource(context):
                return False

        for indirect_perm in self.get_scheme_indirect_perms(context):
            if context.perm not in indirect_perm.can_receive_perms():
                continue

            if indirect_perm.is_authorized_for_unsaved_resource(context):
                return True

        return False

    def get_assigned_perms_q(self, context: Context) -> Optional[Q]:
        q = merge_qs(
            [
                self.get_directly_assigned_perms_q(context),
                self.get_indirectly_assigned_perms_q(context),
            ],
            connector=Q.OR,
        )

        for condition in self.get_scheme_implicit_conditions(context) or []:
            if (condition_q := condition.get_assigned_perms_q(context)) is not None:
                if q is None:
                    q = condition_q
                else:
                    q |= condition_q

        return q

    def get_directly_assigned_perms_q(self, context: Context) -> Optional[Q]:
        if not self.allow_directly_assigned_perms:
            return None

        q = Q(
            user_group__memberships__is_outside_collaborator=False,
            user_group__memberships__user=context.actor,
            perm=context.perm,
        )
        target_resources_q = Q(
            content_type=None, object_id=None
        )  # TODO: is this OK for listing all allowed stuff?

        if context.resource:
            if isinstance(context.resource, Model):
                target_resources_q |= Q(
                    content_type=ContentType.objects.get_for_model(self.model),
                    object_id=context.resource.pk,
                )
            else:
                target_resources_q |= Q(
                    content_type=ContentType.objects.get_for_model(self.model),
                )

        q &= target_resources_q

        return q

    def get_indirectly_assigned_perms_q(self, context: Context) -> Optional[Q]:
        qs = []

        for indirect_perm in self.get_scheme_indirect_perms(context):
            if context.perm not in indirect_perm.can_receive_perms():
                continue

            if (
                indirect_perm_q := indirect_perm.get_assigned_perms_q(context)
            ) is not None:
                qs.append(indirect_perm_q)

        if not qs:
            return None

        return reduce(lambda q1, q2: q1 | q2, qs)

    def get_memberships_q(self, context: Context) -> Optional[Q]:
        q = None
        qs = []

        for indirect_perm in self.get_scheme_indirect_perms(context):
            if context.perm not in indirect_perm.can_receive_perms():
                continue

            if (
                indirect_perm_q := indirect_perm.get_memberships_q(context)
            ) is not None:
                qs.append(indirect_perm_q)

        if qs:
            q = reduce(lambda q1, q2: q1 | q2, qs)

        for condition in self.get_scheme_implicit_conditions(context) or []:
            if (condition_q := condition.get_memberships_q(context)) is not None:
                if q is None:
                    q = condition_q
                else:
                    q |= condition_q

        return q

    def get_auth_scheme_by_relation(self, relation) -> "ModelAuthorizationScheme":
        return get_relation_scheme(self, relation)

    def get_model_for_relation(self, relation) -> type[Model]:
        return self.get_auth_scheme_by_relation(relation).model

    def get_auth_scheme_for_direct_relation(
        self, relation
    ) -> "ModelAuthorizationScheme":
        # TODO: this should raise if there are 2 or more auth classes for the same model
        result = self.auth_solver.get_auth_scheme_for_model(
            self.get_model_for_direct_relation(relation)
        )

        return result

    def get_model_for_direct_relation(self, relation) -> type[Model]:
        result = self.model._meta.get_field(relation).related_model
        if not result:
            raise ValueError(f"No model for relation {relation}")

        return result

    @property
    def model_content_type(self) -> ContentType:
        if not self.model:
            raise ValueError("Missing model.")

        return ContentType.objects.get_for_model(self.model)

    def __str__(self):
        return self.__class__.__name__


def get_relation_scheme(
    initial_scheme: ModelAuthorizationScheme, relation: str
) -> ModelAuthorizationScheme:
    if relation == "*":
        return initial_scheme

    intermediary_relations = relation.split("__")
    related_scheme = initial_scheme

    for relation in intermediary_relations:
        related_scheme = related_scheme.get_auth_scheme_for_direct_relation(relation)

    return related_scheme



#  Copyright 2024 Pressinfra SRL
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations
import enum

from django.core.exceptions import FieldDoesNotExist, ObjectDoesNotExist
from django.db.models import Q, Field
from functools import reduce
from typing import Optional, Callable, TYPE_CHECKING

from django_woah.models import UserGroup
from django_woah.utils.q import (
    get_object_relation,
    merge_qs,
    prefix_q_with_relation,
    verify_resource_by_q,
)
from .context import Context
from .enum import PermEnum

if TYPE_CHECKING:
    from . import ConditionalPerms


class Condition:
    def __init__(self, **kwargs):
        self.scheme = kwargs.get("scheme")

    def get_resources_q(self, context: Context) -> Optional[Q]:
        raise NotImplementedError

    def get_assigned_perms_q(self, context: Context) -> Optional[Q]:
        return None

    def get_memberships_q(self, context: Context) -> Optional[Q]:
        return None

    def is_authorized_for_prefetched_resource(self, context: Context) -> bool:
        return False

    def is_authorized_for_unsaved_resource(self, context: Context) -> bool:
        return False

    def set_scheme(self, scheme):
        if self.scheme is None:
            self.scheme = scheme

    def __and__(self, other):
        return CombinedCondition(
            self, other, operation=CombinedCondition.OPERATIONS.AND
        )

    def __or__(self, other):
        return CombinedCondition(self, other, operation=CombinedCondition.OPERATIONS.OR)

    def __rshift__(self, other: list[PermEnum]) -> "ConditionalPerms":
        from .indirect_perms import ConditionalPerms

        return ConditionalPerms(conditions=[self], receives_perms=other)


class CombinedCondition(Condition):
    class OPERATIONS(enum.StrEnum):
        AND = "and"
        OR = "or"

    def __init__(self, *conditions: Condition, operation: OPERATIONS, **kwargs):
        self.conditions = conditions
        self.operation = operation

        super().__init__(**kwargs)

    def get_resources_q(self, context: Context) -> Optional[Q]:
        result = Q()

        for condition in self.conditions:
            q = condition.get_resources_q(context)

            if self.operation == self.OPERATIONS.AND:
                if q is None:
                    return None

                result &= q
            elif self.operation == self.OPERATIONS.OR:
                if q is not None:
                    result |= q
            else:
                raise ValueError("Unexpected Condition operation")

        return result

    def get_assigned_perms_q(self, context: Context) -> Optional[Q]:
        qs = [
            q
            for condition in self.conditions
            if (q := condition.get_assigned_perms_q(context)) is not None
        ]

        if not qs:
            return None

        return reduce(
            lambda q1, q2: q1 | q2,
            qs,
        )

    def is_authorized_for_prefetched_resource(self, context: Context) -> bool:
        if not self.conditions:
            return True

        for condition in self.conditions:
            ok = condition.is_authorized_for_prefetched_resource(context)

            if self.operation == self.OPERATIONS.AND:
                if not ok:
                    return False
            elif self.operation == self.OPERATIONS.OR:
                if ok:
                    return True
            else:
                raise ValueError("Unexpected Condition operation")

        return self.operation == self.OPERATIONS.AND

    def is_authorized_for_unsaved_resource(self, context: Context) -> bool:
        if not self.conditions:
            return True

        for condition in self.conditions:
            ok = condition.is_authorized_for_unsaved_resource(context)

            if self.operation == self.OPERATIONS.AND:
                if not ok:
                    return False
            elif self.operation == self.OPERATIONS.OR:
                if ok:
                    return True
            else:
                raise ValueError("Unexpected Condition operation")

        return self.operation == self.OPERATIONS.AND

    def set_scheme(self, scheme):
        if self.scheme is None:
            self.scheme = scheme

        for condition in self.conditions:
            condition.set_scheme(self.scheme)


class BaseOwnerCondition(Condition):
    def __init__(self, actor, **kwargs):
        super().__init__(**kwargs)

        self.actor = actor

    def set_scheme(self, scheme):
        super().set_scheme(scheme)

        relation_model = (
            self.scheme.model
            if self.scheme.owner_relation == "*"
            else self.scheme.get_model_for_relation(self.scheme.owner_relation)
        )
        self.relation_is_user_group = relation_model == UserGroup
        self.relation = self.scheme.owner_relation
        self.account_owner_relation = (
            self.relation
            if not self.relation_is_user_group
            else (f"{self.relation}__owner" if not self.relation == "*" else "owner")
        )


class HasRootMembership(BaseOwnerCondition):
    def __init__(self, actor, is_outside_collaborator=None, **kwargs):
        super().__init__(actor=actor, **kwargs)

        self.is_outside_collaborator = is_outside_collaborator

    def get_memberships_q(self, context: Context) -> Q:
        owner = get_object_relation(context.resource, self.relation)

        q = (
            Q(user_group=owner)
            if self.relation_is_user_group
            else Q(user_group__owner=owner)
        )

        q &= Q(user=context.actor)

        if self.is_outside_collaborator is not None:
            q &= Q(is_outside_collaborator=self.is_outside_collaborator)

        return q

    def get_resources_q(self, _: Context) -> Q:
        user_groups_relation = (
            "owned_user_groups__" if not self.relation_is_user_group else ""
        )
        relation = "" if self.relation == "*" else f"{self.relation}__"

        query = {
            f"{relation}{user_groups_relation}memberships__user": self.actor,
            f"{relation}{user_groups_relation}kind": "root",
        }
        if self.is_outside_collaborator is not None:
            query[
                f"{relation}{user_groups_relation}memberships__is_outside_collaborator"
            ] = self.is_outside_collaborator

        return Q(**query)

    def _get_owner(self, resource):
        try:
            return get_object_relation(resource, self.relation)
        except ObjectDoesNotExist:
            return None
        except AttributeError as exception:
            if str(exception).startswith("'NoneType' object has no attribute"):
                return None

            raise  # TODO: Decide if reraise is good here
        except ValueError as exception:
            if str(exception).endswith(
                "before this many-to-many relationship can be used."
            ):
                return None

            raise

    def is_authorized_for_prefetched_resource(self, context: Context) -> bool:
        resource = context.resource
        owner = self._get_owner(resource)

        if not owner:
            return False

        if not self.relation_is_user_group:
            owner = owner.owned_user_groups.get(kind="root")

        if isinstance(owner, UserGroup):
            if not owner.kind == "root":
                return False

        for membership in context.memberships:
            if membership.user_group == owner:
                return True

        return False

    def is_authorized_for_unsaved_resource(self, context: Context) -> bool:
        resource = context.resource
        owner = self._get_owner(resource)

        if not owner:
            return False

        if not self.relation_is_user_group:
            owner = owner.owned_user_groups.get(kind="root")

        if isinstance(owner, UserGroup):
            if not owner.kind == "root":
                return False

        query = {"user": self.actor}

        if self.is_outside_collaborator is not None:
            query["is_outside_collaborator"] = self.is_outside_collaborator

        return owner.memberships.filter(**query).exists()


class HasSameResourcePerms(Condition):
    def __init__(self, perms: list[PermEnum], **kwargs):
        super().__init__(**kwargs)

        self.perms = perms

    def get_resources_q(self, context: Context) -> Optional[Q]:
        return merge_qs(
            [
                self.scheme.get_resources_q(context.subcontext(perm))
                for perm in self.perms
            ]
        )

    def get_assigned_perms_q(self, context: Context) -> Optional[Q]:
        return merge_qs(
            [
                self.scheme.get_assigned_perms_q(context.subcontext(perm))
                for perm in self.perms
            ],
            connector=Q.OR,
        )

    def is_authorized_for_prefetched_resource(self, context: Context) -> bool:
        return all(
            self.scheme.is_authorized_for_prefetched_resource(context.subcontext(perm))
            for perm in self.perms
        )

    def is_authorized_for_unsaved_resource(self, context: Context) -> bool:
        solver = self.scheme.auth_solver

        return all(
            solver.is_authorized_for_unsaved_resource(context.subcontext(perm))
            for perm in self.perms
        )

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.scheme.model} < {self.perms}"


class HasRelatedResourcePerms(Condition):
    def __init__(
        self,
        relation: str,
        perms: list[PermEnum],
        unsaved_object_relation=None,
        **kwargs,
    ):
        if not isinstance(perms, (list, set, tuple)):
            raise ValueError(f"Received perms of type {type(perms)}: {perms}")

        self.perms = perms
        self.relation = relation
        # The "unsaved_object_relation" parameter may be used when dealing with GenericForeignKey relations, because
        # Django doesn't allow assigning to the reverse of a GenericRelation even if that basically means assigning
        # to the GenericForeignKey (the GenericRelation doesn't exist in the DB anyway).
        # For example for "relation" you should set the GenericRelation.related_query_name, but for
        # "unsaved_object_relation" you should set the GenericForeignKey field name (usually "content_object")
        self.unsaved_object_relation = unsaved_object_relation or self.relation
        self.field: Field

        super().__init__(**kwargs)

    def set_scheme(self, scheme):
        super().set_scheme(scheme)

        try:
            self.related_scheme = self.scheme.get_auth_scheme_by_relation(self.relation)
        except ValueError:
            # TODO: only catch this when the relation is generic, because most of the times the ValueError exc is valid
            #  and should raise
            model = self.perms[0].auth_scheme.model
            self.related_scheme = self.scheme.auth_solver.get_auth_scheme_for_model(
                model
            )

        # TODO: check the relationship all the way, not just the first field
        model = self.scheme.model
        field_name = self.relation.split("__", 1)[0]

        try:
            self.field = model._meta.get_field(field_name)
        except FieldDoesNotExist:
            raise AttributeError(
                f"{field_name} was specified in {self.__class__} 'owner_relation', but doesn't exist on {model}"
            )

    def get_resources_q(self, context: Context) -> Optional[Q]:
        qs = [
            prefix_q_with_relation(q, self.relation)
            for perm in self.perms
            if (
                (
                    q := self.related_scheme.get_resources_q(
                        context.subcontext(
                            perm=perm, resource=self.related_scheme.model
                        )
                    )
                )
                is not None
            )
        ]

        if not qs:
            return None

        return reduce(lambda q1, q2: q1 & q2, qs)

    def get_assigned_perms_q(self, context: Context) -> Optional[Q]:
        qs = [
            q
            for perm in self.perms
            if (
                q := self.related_scheme.get_assigned_perms_q(
                    context.subcontext(perm=perm, resource=self.related_scheme.model)
                )
            )
            is not None
        ]
        if not qs:
            return None

        return reduce(
            lambda q1, q2: q1 | q2,
            qs,
        )

    def is_authorized_for_prefetched_resource(self, context: Context) -> bool:
        resource = get_object_relation(context.resource, self.relation)

        if not resource:
            return False

        return all(
            self.related_scheme.is_authorized_for_prefetched_resource(
                context.subcontext(perm, resource)
            )
            for perm in self.perms
        )

    def is_authorized_for_unsaved_resource(self, context: Context) -> bool:
        resource = get_object_relation(context.resource, self.unsaved_object_relation)

        if not resource:
            return False

        solver = self.scheme.auth_solver
        #
        # context.resource = self.scheme.get_auth_scheme_by_relation(
        #     self.relation
        # ).model
        # context.assigned_perms = solver.get_assigned_perms_queryset(context)

        # TODO: filter(pk=context.resource.pk). should be enforced by the solver; remove from here when implemented
        return all(
            solver.get_authorized_on_resources_queryset(
                context.subcontext(perm, resource)
            )
            .filter(pk=resource.pk)
            .exists()
            for perm in self.perms
        )

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.scheme.model}.{self.relation} < {self.perms}"


class HasUnrelatedResourcePerms(Condition):
    def __init__(self, resource, perms: list[PermEnum], **kwargs):
        if not isinstance(perms, (list, set, tuple)):
            raise ValueError(f"Received perms of type {type(perms)}: {perms}")

        self.resource = resource

        self.perms = perms

        super().__init__(**kwargs)

    def get_resources_q(self, context: Context) -> Optional[Q]:
        solver = self.scheme.auth_solver

        # TODO: Optimize below into a single queryset that uses filter(pk=self.resource.pk).exists()
        if all(
            self.resource
            in solver.get_authorized_on_resources_queryset(
                context.subcontext(perm, self.resource)
            )
            for perm in self.perms
        ):
            return Q()

        return None

    def get_assigned_perms_q(self, context: Context) -> Optional[Q]:
        solver = self.scheme.auth_solver

        qs = [
            q
            for perm in self.perms
            if (
                q := solver.get_assigned_perms_q(
                    context.subcontext(perm, self.resource)
                )
            )
            is not None
        ]
        if not qs:
            return None

        return reduce(
            lambda q1, q2: q1 | q2,
            qs,
        )

    def is_authorized_for_prefetched_resource(self, context: Context) -> bool:
        related_scheme = self.scheme.auth_solver.get_auth_scheme_for_model(
            self.resource.__class__
        )

        return all(
            related_scheme.is_authorized_for_prefetched_resource(
                context.subcontext(perm, self.resource)
            )
            for perm in self.perms
        )

    def is_authorized_for_unsaved_resource(self, context: Context) -> bool:
        # TODO: Implement
        return self.get_resources_q(context) == Q()

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.resource} < {self.perms}"


class QCondition(Condition):
    def __init__(
        self, q: Q, authorize_unsaved_resource_func: Optional[Callable] = None, **kwargs
    ):
        self.q = q
        self.authorize_unsaved_resource_func = authorize_unsaved_resource_func

        super().__init__(**kwargs)

    def get_resources_q(self, _: Context) -> Q:
        return self.q

    def is_authorized_for_prefetched_resource(self, context: Context) -> bool:
        return verify_resource_by_q(context.resource, self.q)

    def is_authorized_for_unsaved_resource(self, context: Context) -> bool:
        if self.authorize_unsaved_resource_func:
            return self.authorize_unsaved_resource_func(context)

        return verify_resource_by_q(context.resource, self.q)

    def __repr__(self):
        return self.q.__repr__()



#  Copyright 2024 Pressinfra SRL
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
from django.db.models import Q
from sympy import And, Or, Not, simplify_logic, Symbol
from sympy.logic.boolalg import BooleanTrue, BooleanFalse
from typing import Optional


def merge_qs(qs: list[Optional[Q]], connector: str = Q.AND) -> Optional[Q]:
    result: Optional[Q] = None

    for q in qs:
        if q is None:
            if connector == Q.AND:
                return None

            continue

        result = result._combine(q, conn=connector) if result is not None else q

    return result


def prefix_q_with_relation(q: Q, relation: str) -> Q:
    children = []

    for child in q.children:
        if isinstance(child, Q):
            children.append(prefix_q_with_relation(child, relation))
        elif isinstance(child, tuple):
            prefixed_relation = (
                f"{relation}__{child[0]}" if child[0] != "*" else relation
            )

            children.append((prefixed_relation, *child[1:]))
        else:
            raise ValueError("Unexpected child", child)

    return Q(*children, _connector=q.connector, _negated=q.negated)


def pop_parts_of_q(q: Q, matcher) -> Q:
    children = []

    for child in q.children:
        if isinstance(child, Q):
            children.append(pop_parts_of_q(child, matcher))
        elif isinstance(child, tuple):
            if matcher(*child):
                continue

            children.append((child[0], *child[1:]))
        else:
            raise ValueError("Unexpected child", child)

    return Q(*children, _connector=q.connector, _negated=q.negated)


def remove_prefix_relation_from_q(q: Q, relation: str) -> Q:
    children = []
    prefix = f"{relation}__"

    for child in q.children:
        if isinstance(child, Q):
            children.append(remove_prefix_relation_from_q(child, relation))
        elif isinstance(child, tuple):
            stripped_relation = child[0]

            if stripped_relation.startswith(prefix):
                stripped_relation = stripped_relation[len(prefix) :]

            if not stripped_relation:
                stripped_relation = "*"

            children.append((stripped_relation, *child[1:]))
        else:
            raise ValueError("Unexpected child", child)

    return Q(*children, _connector=q.connector, _negated=q.negated)


def get_object_relation(obj, relation: str):
    if relation == "*":
        return obj

    for intermediary_attr in relation.split("__"):
        obj = getattr(obj, intermediary_attr)

    return obj


def verify_resource_by_q(resource, q: Q) -> bool:
    result = True

    for child in q.children:
        if q.connector == Q.OR:
            operator = result.__or__
        elif q.connector == Q.AND:
            operator = result.__and__
        else:
            raise ValueError("Unexpected Q operator", q.connector)

        if isinstance(child, Q):
            result = operator(verify_resource_by_q(resource, child))
        elif isinstance(child, tuple):
            return get_object_relation(resource, child[0]) == child[1]
        else:
            raise ValueError("Unexpected child", child)

    return result


def _prepare_tuple(t: tuple) -> tuple:
    key, value = t
    if isinstance(value, list):
        value = frozenset(value)

    if key == "*":
        return "pk", value.pk

    key = key.replace("__*", "").replace("*__", "")

    return key, value


whitespace_re = re.compile(r"\s+")


def get_sympy_symbols_to_q_children(q: Q, symbols_table: Optional[dict] = None) -> dict:
    if symbols_table is None:
        symbols_table = {}

    for child in q.children:
        if isinstance(child, Q):
            get_sympy_symbols_to_q_children(child, symbols_table)

        else:
            key = f"{child[0]}={re.sub(whitespace_re, '', child[1].__repr__())}"
            while (value := symbols_table.get(key)) is not None and value != child:
                key += "(1)"

            symbols_table[Symbol(key)] = child

    return symbols_table


def get_expression(q: Q, q_children_to_sympy_symbols: dict):
    expression_node = And if q.connector == q.AND else Or

    expressions = []

    for child in q.children:
        if isinstance(child, Q):
            expressions.append(get_expression(child, q_children_to_sympy_symbols))
        else:
            expressions.append(q_children_to_sympy_symbols[child])

    result = expression_node(*expressions)
    if q.negated:
        result = Not(result)

    return result


def _prepare_q(q: Q) -> Q:
    return Q(
        *set(
            dict.fromkeys(
                (
                    _prepare_tuple(child)
                    if not isinstance(child, Q)
                    else _prepare_q(child)
                )
                for child in q.children
            )
        ),
        _connector=q.connector,
        _negated=q.negated,
    )


def _sympy_expression_to_q(
    expression, sympy_symbols_to_q_children: dict, _toplevel=False
) -> bool | Q:
    if isinstance(expression, (bool, BooleanTrue, BooleanFalse)):
        return bool(expression)

    if isinstance(expression, Not):
        # Theoretically not possible to have not with more than 1 arg
        arg = _sympy_expression_to_q(expression.args[0], sympy_symbols_to_q_children)
        if isinstance(arg, Q):
            arg.negated = not arg.negated

            return arg

        else:
            return not arg

    if isinstance(expression, Symbol):
        return (
            sympy_symbols_to_q_children[expression]
            if not _toplevel
            else Q(sympy_symbols_to_q_children[expression])
        )

    connector = Q.AND if isinstance(expression, And) else Q.OR

    return Q(
        *[
            _sympy_expression_to_q(arg, sympy_symbols_to_q_children)
            for arg in expression.args
        ],
        _connector=connector,
    )


def sympy_expression_to_q(
    expression, sympy_symbols_to_q_children: dict, _toplevel=False
) -> bool | Q:
    return _sympy_expression_to_q(
        expression, sympy_symbols_to_q_children, _toplevel=True
    )


def optimize_q(q: Optional[Q], allow_bools=True) -> Q | bool | None:
    if not q:
        return None

    q = _prepare_q(q)

    sympy_to_q_children = get_sympy_symbols_to_q_children(q)
    q_children_to_sympy = {value: key for key, value in sympy_to_q_children.items()}

    expression = get_expression(q, q_children_to_sympy_symbols=q_children_to_sympy)
    simplified_expression = simplify_logic(expression)

    result = sympy_expression_to_q(simplified_expression, sympy_to_q_children)

    if not allow_bools:
        if result is False:
            return None

        if q is True:
            return Q()

    return result



#  Copyright 2024 Pressinfra SRL
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass, field
from django.contrib.auth.models import AbstractUser
from django.db.models import Model
from typing import Optional, Union

from django_woah.models import AssignedPerm, Membership
from .enum import PermEnum


@dataclass
class Context:
    actor: Optional[AbstractUser] = None
    perm: Optional[PermEnum] = None
    resource: Optional[Model | list[Model]] = None
    extra: dict = field(default_factory=dict)

    _assigned_perms: Optional[list[AssignedPerm]] = None
    _memberships: Optional[list[Membership]] = None
    _depth: int = 0
    _root: Optional[Union["Context", "CombinedContext"]] = None

    def subcontext(self, perm=None, resource=None):
        return Context(
            actor=self.actor,
            perm=perm or self.perm,
            resource=resource or self.resource,
            extra=self.extra,
            # _assigned_perms=self.assigned_perms,
            _depth=self._depth,
            _root=self._root or self,
        )

    @property
    def assigned_perms(self):
        if self._assigned_perms is None and self._root:
            return self._root.assigned_perms

        return self._assigned_perms

    @assigned_perms.setter
    def assigned_perms(self, value):
        self._assigned_perms = value

    @property
    def memberships(self):
        if self._memberships is None and self._root:
            return self._root._memberships

        return self._memberships

    @memberships.setter
    def memberships(self, value):
        self._memberships = value


@dataclass
class CombinedContext:
    contexts: list[Context] = field(default_factory=list)
    assigned_perms: Optional[list[AssignedPerm]] = None
    memberships: Optional[list[Membership]] = None

    def __post_init__(self):
        self.validate()

    def add(self, context: Context):
        if not context._root:
            context._root = self

        self.contexts.append(context)
        self.validate()

    def validate(self):
        if not self.contexts:
            return

        actor = self.contexts[0].actor
        resource = self.contexts[0].resource

        for context in self.contexts[1:]:
            if context.actor != actor:
                raise ValueError("Cannot have contexts with different actors.")

            if context.resource != resource:
                raise ValueError("Cannot have contexts with different resources.")



#  Copyright 2024 Pressinfra SRL
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from django.db import models


class AutoCleanModel(models.Model):
    class Meta:
        abstract = True

    def _init_states(self):
        self.initial_state = self.current_state

        self.cleaned_state = {} if not self.pk else self.initial_state.copy()
        self.saved_state = {} if not self.pk else self.initial_state.copy()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._init_states()

    @property
    def current_state(self):
        return {
            field.name: self.__dict__[field.attname]
            for field in self._meta.fields
            if field.attname in self.__dict__
        }

    @staticmethod
    def _states_diff(state, other_state):
        return {key: value for key, value in other_state.items() if value != state[key]}

    def get_dirty_fields(self):
        return self._states_diff(self.current_state, self.cleaned_state)

    def get_unsaved_fields(self):
        if not self.saved_state:
            return list(self.current_state.keys())

        return list(self._states_diff(self.current_state, self.saved_state).keys())

    @property
    def is_cleaned(self):
        if not getattr(self, ".cleaned", False):
            return False

        return not self.get_dirty_fields()

    @is_cleaned.setter
    def is_cleaned(self, value):
        if value:
            self.cleaned_state = self.current_state.copy()

        setattr(self, ".cleaned", value)

    def save(self, *args, **kwargs):
        if not self.is_cleaned:
            self.full_clean()

        super().save(*args, **kwargs)

        self.initial_state = self.current_state.copy()
        if kwargs.get("update_fields") is None:
            self.saved_state = self.current_state.copy()
        else:
            for field in kwargs["update_fields"]:
                if field not in self.current_state:
                    continue

                self.saved_state[field] = self.current_state[field]

    def refresh_from_db(self, *args, **kwargs):
        super().refresh_from_db(*args, **kwargs)

        self._init_states()

    def full_clean(self, *args, **kwargs):
        if self.is_cleaned:
            return

        super().full_clean(*args, **kwargs)

        self.is_cleaned = True



#  Copyright 2024 Pressinfra SRL
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from inspect import isclass
from typing import Optional

from django.contrib.auth import get_user_model
from django.db.models import Q, Model, Manager, Subquery

from django_woah.models import AssignedPerm, Membership
from django_woah.utils.q import merge_qs, optimize_q, pop_parts_of_q
from .context import Context, CombinedContext
from .enum import PermEnum
from .scheme import ModelAuthorizationScheme


class AuthorizationSolver:
    def __init__(
        self,
        authorization_schemes: list[
            ModelAuthorizationScheme | type[ModelAuthorizationScheme]
        ],
        clean_perms=True,
    ):
        self.authorization_schemes: list[ModelAuthorizationScheme] = []

        for scheme in authorization_schemes:
            if not isinstance(scheme, ModelAuthorizationScheme):
                scheme = scheme()

            scheme.auth_solver = self
            self.authorization_schemes.append(scheme)

        if clean_perms:
            perms = {}

            for scheme in self.authorization_schemes:
                for perm in scheme.get_scheme_perms(exclude_borrowed=True):
                    if perm.value in perms:
                        raise ValueError(
                            f"Found {perm.value} in both {perm.auth_scheme} and {perms[perm.value].auth_scheme}"
                        )

                    perms[perm.value] = perm

    def clean_perm(
        self, dirty_perm: str | PermEnum
    ) -> tuple[PermEnum, ModelAuthorizationScheme]:
        for scheme in self.authorization_schemes:
            for perm in getattr(scheme, "Perms", []):
                if dirty_perm == perm:
                    return perm, scheme

            for perm in getattr(scheme, "Roles", []):
                if dirty_perm == perm:
                    return perm, scheme

        raise ValueError(f"Unexpected perm received: {dirty_perm}")

    def get_assigned_perms_q(self, context: Context) -> Optional[Q]:
        _, scheme = self.clean_perm(context.perm)

        q = scheme.get_assigned_perms_q(context)

        return q

    def get_memberships_q(self, context: Context) -> Optional[Q]:
        _, scheme = self.clean_perm(context.perm)

        q = scheme.get_memberships_q(context)

        return q

    def get_assigned_perms_queryset(
        self,
        root_context: Context | CombinedContext,
    ):
        contexts = (
            root_context.contexts
            if isinstance(root_context, CombinedContext)
            else [root_context]
        )

        assigned_perms_q = merge_qs(
            [self.get_assigned_perms_q(context) for context in contexts], connector=Q.OR
        )

        if assigned_perms_q is None:
            return AssignedPerm.objects.none()

        return AssignedPerm.objects.filter(assigned_perms_q)

    def get_context(
        self,
        actor,
        perm: str | PermEnum,
        resource=None,
        prefetch_assigned_perms=True,
        extra: Optional[dict] = None,
        **kwargs,
    ) -> Context:
        if isinstance(perm, PermEnum):
            perm, _ = self.clean_perm(perm)

        if extra is None:
            extra = {}

        context = Context(
            actor=actor, perm=perm, resource=resource, extra=extra, **kwargs
        )

        if prefetch_assigned_perms and "assigned_perms" not in kwargs:
            context.assigned_perms = self.get_assigned_perms_queryset(context)

        return context

    def get_context_no_actor(
        self,
        perm: str | PermEnum,
        resource=None,
        prefetch_assigned_perms=True,
        prefetch_memberships=True,
        extra: Optional[dict] = None,
        **kwargs,
    ) -> Context:
        if isinstance(perm, PermEnum):
            perm, _ = self.clean_perm(perm)

        if extra is None:
            extra = {}

        # Use a Fake actor to get assigned_perms_q and memberships
        class FakePK(int):
            def is_fake(self):
                return True

        context = Context(
            actor=get_user_model()(pk=FakePK(-1)),
            perm=perm,
            resource=resource,
            extra=extra,
            **kwargs,
        )

        # Strip the Fake actor from the assigned perms Q to fetch all the potentially needed ones
        assigned_perms_q = pop_parts_of_q(
            self.get_assigned_perms_q(context),
            matcher=lambda key, *values: key == "user_group__memberships__user"
            and values[0] == context.actor,
        )

        # TODO: See how to extract this into a separate method, maybe the existing get_assigned_perms_queryset
        if prefetch_assigned_perms and "assigned_perms" not in kwargs:
            context.assigned_perms = AssignedPerm.objects.filter(assigned_perms_q)

        # TODO: See how to extract this into a separate method
        if prefetch_memberships and "memberships" not in kwargs:
            memberships_q = Q(
                user_group__in=[
                    assigned_perm.user_group_id
                    for assigned_perm in context.assigned_perms
                ]
            )

            if (additional_q := self.get_memberships_q(context)) is not None:
                # Strip the fake actor from the memberships Q to fetch all the potentially needed ones
                additional_q = pop_parts_of_q(
                    additional_q,
                    matcher=lambda key, *values: key == "user"
                    and values[0] == context.actor,
                )

                memberships_q |= additional_q

            context.memberships = Membership.objects.filter(memberships_q)

        context.actor = None

        return context

    def get_auth_scheme_for_model(self, model: type[Model]):
        for scheme in self.authorization_schemes:
            if scheme.model == model:
                return scheme

        raise ValueError(f"No AuthScheme for model {model}")

    def get_model(self, resources):
        if isinstance(resources, Model):
            model = resources.__class__
        elif isinstance(resources, Manager):
            model = resources.model
        elif issubclass(resources, Model):
            model = resources
        else:
            model = resources

            if not isclass(model):
                model = model[0].__class__

        return model

    def get_authorized_actors_q(self, context: Context) -> Optional[Q]:
        if context.actor is not None:
            raise ValueError("Must not specify context actor")

        if not context.resource:
            raise ValueError("Must specify context resource")

        if context.assigned_perms is None:
            raise ValueError("Must specify context assigned_perms")

        if context.memberships is None:
            raise ValueError("Must specify context memberships")

        model = self.get_model(context.resource)
        scheme = self.get_auth_scheme_for_model(model)

        assigned_perms_by_user_group_id = {}

        for assigned_perm in context.assigned_perms:
            if assigned_perm.user_group_id not in assigned_perms_by_user_group_id:
                assigned_perms_by_user_group_id[assigned_perm.user_group_id] = [
                    assigned_perm
                ]
            else:
                assigned_perms_by_user_group_id[assigned_perm.user_group_id].append(
                    assigned_perm
                )

        actors_context_data = {}

        for membership in context.memberships:
            actor_id = membership.user_id

            if not actors_context_data.get(actor_id):
                actors_context_data[actor_id] = {
                    "memberships": [membership],
                    "assigned_perms": assigned_perms_by_user_group_id.get(
                        membership.user_group_id, []
                    ),
                }
            else:
                actors_context_data[actor_id]["memberships"].append(membership)
                actors_context_data[actor_id]["assigned_perms"] += (
                    assigned_perms_by_user_group_id.get(membership.user_group_id, [])
                )

        actor_class = get_user_model()
        authorized_actors_ids = []

        for actor_id, context_data in actors_context_data.items():
            subcontext = context.subcontext()
            subcontext.actor = actor_class(id=actor_id)
            subcontext.assigned_perms = context_data["assigned_perms"]
            subcontext.memberships = context_data["memberships"]

            if scheme.is_authorized_for_prefetched_resource(subcontext):
                authorized_actors_ids.append(actor_id)

        if not authorized_actors_ids:
            return None

        return Q(id__in=authorized_actors_ids)

    def get_authorized_actors_queryset(self, context: Context):
        actor_class = get_user_model()

        q = self.get_authorized_actors_q(context)
        if q is None:
            return actor_class.objects.none()

        return actor_class.objects.filter(q)

    def get_authorized_on_resources_q(self, context: Context) -> Optional[Q]:
        if not context.resource:
            raise ValueError("Must specify context resource")

        if context.assigned_perms is None:
            raise ValueError("Must specify context assigned_perms")

        model = self.get_model(context.resource)
        scheme = self.get_auth_scheme_for_model(model)

        resources_q = scheme.get_resources_q(context)
        if resources_q is None:
            return None

        return optimize_q(resources_q, allow_bools=False)

    def get_authorized_on_resources_queryset(
        self,
        context: Optional[Context | CombinedContext] = None,
        base_queryset=None,
        **kwargs,
    ):
        if not context:
            context = self.get_context(**kwargs)

        if isinstance(context, CombinedContext):
            contexts = context.contexts
        else:
            contexts = [context]

        model = self.get_model(contexts[0].resource)

        q = merge_qs(
            [self.get_authorized_on_resources_q(context) for context in contexts]
        )

        if q is None:
            return model.objects.none()

        if base_queryset is None:
            base_queryset = model.objects.all()

        # Copy the base_queryset, detach it's ordering, and attach it to the new queryset
        order_by = base_queryset.query.order_by
        extra_order_by = base_queryset.query.extra_order_by
        default_ordering = base_queryset.query.default_ordering

        base_queryset.query.clear_ordering()

        queryset = model.objects.filter(
            pk__in=Subquery(base_queryset.filter(q).values("pk").distinct())
        )

        queryset.query.order_by = order_by
        queryset.query.extra_order_by = extra_order_by
        queryset.query.default_ordering = default_ordering

        return queryset

    def is_authorized_for_prefetched_resource(
        self, context: Optional[Context] = None, **kwargs
    ) -> bool:
        if not context:
            context = self.get_context(**kwargs)

        if not context.resource:
            raise ValueError("Must specify resource")

        if not isinstance(context.resource, Model):
            raise ValueError(
                f"Expected resource to be a Model instance, but got: {context.resource}"
            )

        model = context.resource.__class__
        scheme = self.get_auth_scheme_for_model(model)

        return scheme.is_authorized_for_prefetched_resource(context)

    def is_authorized_for_unsaved_resource(
        self, context: Optional[Context | CombinedContext] = None, **kwargs
    ) -> bool:
        if not context:
            context = self.get_context(**kwargs)

        if isinstance(context, CombinedContext):
            contexts = context.contexts
        else:
            contexts = [context]

        for context in contexts:
            if not context.resource:
                raise ValueError("Must specify resource")

            if not isinstance(context.resource, Model):
                raise ValueError(
                    f"Expected resource to be a Model instance, but got: {context.resource}"
                )

            model = context.resource.__class__
            scheme = self.get_auth_scheme_for_model(model)

            if not scheme.is_authorized_for_unsaved_resource(context):
                return False

        return True


# def gather_schemes():
#     sub_schemes = AuthorizationScheme.__subclasses__()
#     leaf_schemes = []
#
#     for scheme in sub_schemes:
#         if not scheme.__subclasses__():
#             leaf_schemes.append(scheme)
#         else:
#             leaf_schemes.extend(gather_schemes(scheme))
#
#     return leaf_schemes
#
# DefaultAuthorizationSolver = AuthorizationSolver(
#     authorization_schemes=gather_schemes()
# )



#  Copyright 2024 Pressinfra SRL
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections import defaultdict

from django.conf import settings
from django.core.exceptions import ValidationError as DjangoValidationError
from django.db.models import Model
from django.db.models import Q
from django.http import Http404
from rest_framework import mixins
from rest_framework.exceptions import (
    MethodNotAllowed,
    PermissionDenied,
    ValidationError,
)
from rest_framework.fields import empty
from rest_framework.permissions import IsAuthenticated
from rest_framework.viewsets import GenericViewSet
from typing import Optional

from django_woah.authorization import PermEnum
from django_woah.authorization.context import CombinedContext
from django_woah.authorization.solver import AuthorizationSolver
from django_woah.drf.permission import IsAuthorized
from django_woah.utils.q import merge_qs

uninitialized = object()

validation_error_setting = "AUTHORIZATION_UNSAVED_RESOURCE_VALIDATION_ERRORS"
clean_unsaved_resource_setting = "AUTHORIZATION_UNSAVED_RESOURCE_CLEAN_BEFORE"
validation_error_message = (
    "The unsaved resource on which the authorization is being run against has validation errors.\n"
    f"To easily avoid most of these situations, you may set {validation_error_setting} = False"
    "in your app settings, but beware of the (rare) potential pitfalls.\n"
    "Check the base `get_unsaved_resource` method on your view to find more about it."
)


class AuthorizationViewSetMixin:
    authorization_solver: AuthorizationSolver

    permission_classes = [IsAuthenticated, IsAuthorized]

    def __init__(self, *args, **kwargs):
        self._cache = defaultdict(lambda: uninitialized, {})

        super().__init__(*args, **kwargs)

    cache_separator = object()

    def get_cache_key(self, key, *args, **kwargs):
        return (
            (key,)
            + (self.cache_separator,)
            + args
            + (self.cache_separator,)
            + tuple(sorted(kwargs.items()))
        )

    @property
    def model(self) -> type[Model]:
        if queryset := getattr(self, "queryset", None):
            return queryset.model

        if serializer_class := self.get_serializer_class():
            return serializer_class.Meta.model

        raise ValueError("A model property must be defined on your view.")

    def get_authorization_relation(self):
        return getattr(self, "authorization_relation", None)

    def get_authorized_model_lookup_url_kwarg(self):
        return getattr(
            self,
            "authorized_model_lookup_url_kwarg",
            getattr(self, "lookup_url_kwarg", None),
        )

    def get_authorized_model_lookup_field(self):
        return getattr(
            self, "authorized_model_lookup_field", getattr(self, "lookup_field", None)
        )

    @property
    def authorization_model(self) -> type[Model]:
        auth_relation_field = self.get_authorization_relation()
        if not auth_relation_field:
            return self.model

        return self.model._meta.get_field(auth_relation_field).remote_field.model

    def get_required_permissions(self) -> list[PermEnum]:
        """
        A perms_map in the style of DRF's `DjangoModelPermissions` can be defined on the view.
        Here's a random example:

        perms_map = {
            'GET': [XYZScheme.Perms.ViewXYZ],
            'OPTIONS': [XYZScheme.Perms.ViewXYZ],
            'HEAD': [XYZScheme.Perms.ViewXYZ],
            'POST': [XYZScheme.Perms.CreateXYZ],
            'PUT': [XYZScheme.Perms.EditXYZ],
            'PATCH': [XYZScheme.Perms.EditXYZ],
            'DELETE': [XYZScheme.Perms.ViewXYZ, XYZScheme.Perms.DeleteXYZ],
        }
        """

        if not hasattr(self, "perms_map"):
            raise NotImplementedError(
                "Provide a `perms_map` dict or implement the `get_required_permissions` method in the view."
            )

        method = self.request.method

        perms = self.perms_map.get(method)
        if perms is None:
            if method == "OPTIONS":
                if getattr(self, "authorize_options_as_get", True):
                    perms = self.perms_map.get("GET")

        if perms is None:
            raise MethodNotAllowed(method)

        if isinstance(perms, (str, PermEnum)):
            perms = [perms]

        return perms

    def get_actor(self):
        return self.request.user

    def get_authorization_context_extra(self, perm: str | PermEnum) -> dict:
        return {}

    def get_authorization_context(self) -> CombinedContext:
        cache_key = self.get_cache_key("get_authorization_context")
        cached_result = self._cache[cache_key]

        if cached_result is not uninitialized:
            return cached_result

        context = CombinedContext()

        for perm in self.get_required_permissions():
            if not isinstance(perm, PermEnum):
                perm, _ = self.authorization_solver.clean_perm(perm)

            context.add(
                self.authorization_solver.get_context(
                    actor=self.get_actor(),
                    perm=perm,
                    resource=self.authorization_model,
                    extra=self.get_authorization_context_extra(perm),
                    prefetch_assigned_perms=False,
                )
            )

        context.assigned_perms = self.authorization_solver.get_assigned_perms_queryset(
            context
        )

        self._cache[cache_key] = context

        return context

    def get_authorization_model_q(self) -> Optional[Q]:
        cache_key = self.get_cache_key("get_authorization_model_q")
        cached_result = self._cache[cache_key]

        if cached_result is not uninitialized:
            return cached_result

        qs = []

        for context in self.get_authorization_context().contexts:
            qs.append(
                self.authorization_solver.get_authorized_on_resources_q(context=context)
            )

        q = merge_qs(qs)

        if q is not None and self.kwargs.get(
            self.get_authorized_model_lookup_url_kwarg()
        ):
            q &= Q(
                **{
                    self.get_authorized_model_lookup_field(): self.kwargs.get(
                        self.get_authorized_model_lookup_url_kwarg()
                    )
                }
            )

        self._cache[cache_key] = q

        return q

    def get_authorization_model_queryset(self, base_queryset=None):
        cache_key = self.get_cache_key(
            "get_authorization_model_queryset", base_queryset=base_queryset
        )
        cached_result = self._cache[cache_key]

        if cached_result is not uninitialized:
            return cached_result

        queryset = self.authorization_solver.get_authorized_on_resources_queryset(
            context=self.get_authorization_context(),
            base_queryset=base_queryset,
        )

        if self.kwargs.get(self.get_authorized_model_lookup_url_kwarg()):
            queryset = queryset.filter(
                **{
                    self.get_authorized_model_lookup_field(): self.kwargs.get(
                        self.get_authorized_model_lookup_url_kwarg()
                    )
                }
            )

        self._cache[cache_key] = queryset

        return queryset

    def get_authorization_model_object(
        self, skip_authorization=False
    ) -> Optional[Model]:
        cache_key = self.get_cache_key(
            "get_authorization_model_object", skip_authorization=skip_authorization
        )
        cached_result = self._cache[cache_key]

        if cached_result is not uninitialized:
            return cached_result

        lookup_url_kwarg = self.get_authorized_model_lookup_url_kwarg()
        if lookup_url_kwarg and self.kwargs.get(lookup_url_kwarg) is None:
            return None

        queryset = self.authorization_model.objects.filter(
            **{
                self.get_authorized_model_lookup_field(): self.kwargs.get(
                    lookup_url_kwarg
                )
            }
        )
        if not skip_authorization:
            queryset = self.get_authorization_model_queryset(base_queryset=queryset)

        obj = queryset.first()

        self._cache[cache_key] = obj

        return obj

    def get_requested_model_queryset(self):
        queryset = getattr(self, "queryset", None)

        if self.model == self.authorization_model:
            return self.get_authorization_model_queryset(base_queryset=queryset)

        if not queryset:
            queryset = self.model.objects

        return queryset.filter(
            **{
                f"{self.get_authorization_relation()}__in": self.get_authorization_model_queryset()
            }
        )

    def get_unsaved_resource(self, initial_obj=None) -> Model:
        """
        This method is used to obtain an instance of the unsaved resource against which authorization will be run.
        It is written such that it minimizes the cases where exceptions are raised, in case of ValidationErrors,
        or fields which don't exist on the model or are reverse relations and cannot be instantiated as such.

        It's possible that in some weird and rare cases, this will result in a wrong authorization decision.

        For example:
        - when reverse relations should be part of the validation, because those are not handled here.
            - using the `unsaved_object_relation` field, in the AuthorizationScheme, might help in some of these cases
        - when certain fields fail validation (even if that case should result in a validation error eventually).
        """

        raise_exception = getattr(settings, validation_error_setting, True)
        raised = False

        serializer = self.get_serializer(data=self.request.data)
        model = serializer.Meta.model

        reverse_relations = [
            f.name
            for f in model._meta.get_fields()
            if f.auto_created and not f.concrete
        ]

        try:
            serializer.is_valid(raise_exception=True)
            data = serializer.validated_data
        except (ValidationError, DjangoValidationError):
            if raise_exception:
                raised = True
                raise

            serializer_fields = serializer.fields
            data = {}

            for field_name, field in serializer_fields.items():
                if field.source == "*" or field.read_only:
                    continue

                if (field_value := field.get_value(self.request.data)) != empty:
                    try:
                        data[field.source] = field.to_internal_value(field_value)
                    except (ValidationError, DjangoValidationError):
                        continue
        finally:
            if raised:
                print(validation_error_message)

        data = {k: v for k, v in data.items() if k not in reverse_relations}

        if not initial_obj:
            resource = serializer.Meta.model(**data)
        else:
            resource = initial_obj

            for k, v in data.items():
                setattr(resource, k, v)

        if getattr(
            self,
            "authorization_clean_unsaved_resource",
            getattr(settings, clean_unsaved_resource_setting, True),
        ):
            try:
                resource.clean()
            except DjangoValidationError:
                if raise_exception:
                    raised = True
                    raise
            finally:
                if raised:
                    print(validation_error_message)

        return resource

    def is_authorized_for_unsaved_resource(self) -> bool:
        # TODO move this in context initialization
        combined_context = self.get_authorization_context()

        try:
            authorization_model = self.authorization_model
        except AttributeError:
            authorization_model = None

        try:
            resource_model = self.model
        except AttributeError:
            resource_model = None

        if authorization_model != resource_model:
            # If the authorization model is different from the one being serialized it makes no sense
            # to use get_unsaved_resource, so get_authorization_model_object is used as it's probably
            # the one deciding the authorization

            # TODO: enable this optimization, but only when lookup_field == model.pk field
            # if lookup_value := self.kwargs.get(
            #     self.get_authorized_model_lookup_url_kwarg()
            # ):
            #     context.resource = self.authorization_model(
            #         **{self.get_authorized_model_lookup_field(): lookup_value}
            #     )
            # else:
            resource = self.get_authorization_model_object(skip_authorization=True)
            for context in combined_context.contexts:
                context.resource = resource

            return self.authorization_solver.get_authorized_on_resources_queryset(
                context=combined_context,
            ).exists()

        try:
            resource = self.get_unsaved_resource()

            if resource is None:
                return False

            for context in combined_context.contexts:
                context.resource = resource

            return self.authorization_solver.is_authorized_for_unsaved_resource(
                context=combined_context,
            )
        except (AttributeError, ValueError):
            if not self.request.data:
                # DRF has this "nice" way of testing if it should render a POST form in the default browsable API
                serializer = self.get_serializer(data=self.request.data)

                if not serializer.is_valid(raise_exception=False):
                    # If the payload is not valid, it means we can allow authorization to pass since it will fail
                    # at the validation step even if it were a real request
                    # This will allow a POST form to be rendered in the browsable API, instead of getting a 500 error
                    return True

            raise


class AuthorizationGenericViewSet(AuthorizationViewSetMixin, GenericViewSet):
    def get_queryset(self):
        return self.get_requested_model_queryset()

    def get_object(self):
        try:
            return super().get_object()
        except Http404:
            raise PermissionDenied()


class AuthorizationModelViewSet(
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.UpdateModelMixin,
    mixins.DestroyModelMixin,
    mixins.ListModelMixin,
    AuthorizationGenericViewSet,
):
    pass



#  Copyright 2024 Pressinfra SRL
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import enum


class PermEnum(enum.StrEnum):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        enum.unique(self.__class__)

    def __repr__(self):
        if not hasattr(self, "auth_scheme"):
            return super().__repr__()

        value_repr = self.__class__._value_repr_ or repr

        return "<%s.%s.%s: %s>" % (
            self.auth_scheme.__name__,
            self.__class__.__name__,
            self._name_,
            value_repr(self._value_),
        )

    @classmethod
    def values(cls):
        return list(cls)



from django.db.models import Q

from django_woah.utils.q import optimize_q


def test_optimize_q_1level_simple():
    assert optimize_q(
        Q(Q(a=2), Q(b=3), Q(a=2)),
    ) == Q(a=2) & Q(b=3)


def test_optimize_q_1level_simple_case2():
    assert optimize_q(
        Q(a=2, d=4, b=2) | Q(a=2),
    ) == Q(a=2)


def test_optimize_q_3level_simple_case3():
    # This tests for sorting as well
    assert optimize_q(
        Q(Q(Q(Q(Q(a=2))), Q(b=3)), Q(b=3)) & Q(Q(Q(a=2))),
    ) == Q(a=2, b=3)


def test_optimize_q_2level_simple():
    assert optimize_q(
        Q(
            Q(a=2),
            Q(a=2),
            Q(b=3),
        )
        & Q(b=3)
        & Q(a=2)
    ) == Q(a=2) & Q(b=3)


def test_optimize_q_1level_extract_common():
    assert optimize_q(
        (Q(a=2) & Q(b=2)) | (Q(a=2) & Q(c=3)),
    ) == Q(a=2) & (Q(b=2) | Q(c=3))


def test_optimize_q_1level_extract_common_and_prune_extras():
    assert optimize_q(
        (Q(a=2) & Q(b=2)) & (Q(a=2) & Q(c=3)) & (Q(d=4) | Q(a=2)),
    ) == Q(a=2, b=2, c=3)


def test_optimize_q_1level_extract_common_case2():
    assert optimize_q(
        (Q(a=2) & Q(b=2, a=2)) | (Q(a=2) & Q(c=3)) | Q(a=2, d=4),
    ) == Q(a=2) & (Q(b=2) | Q(c=3) | Q(d=4))


def test_optimize_q_1level_prune_extras():
    assert optimize_q(
        (Q(a=2) & Q(b=2, a=2)) | (Q(a=2, b=2) & Q(c=3)) | (Q(a=2, d=4, b=2) | Q(a=2)),
    ) == Q(a=2)


def test_optimize_q_1level_partially_extract_common():
    assert optimize_q(Q(a=2, b=2) & Q(c=3) & Q(Q(b=2), a=2, c=3)) == Q(a=2, b=2, c=3)


def test_optimize_q_negated_1level_simple():
    assert optimize_q(Q(a=2) & ~Q(a=2)) is False



# Generated by Django 4.2.9 on 2024-02-01 10:57

import django.db.models.deletion
import uuid6
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Membership",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid6.uuid7,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                    ),
                ),
                (
                    "is_outside_collaborator",
                    models.BooleanField(
                        default=False,
                        help_text="Outside collaborators might receive different permissions than members, or even no permissions.",
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_woah.membership",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="UserGroup",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid6.uuid7,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                    ),
                ),
                ("name", models.CharField(blank=True, max_length=80)),
                (
                    "kind",
                    models.CharField(
                        choices=[("root", "Root"), ("team", "Team"), ("user", "User")],
                        max_length=16,
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="owned_user_groups",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="children",
                        to="django_woah.usergroup",
                    ),
                ),
                (
                    "parent_membership",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_woah.membership",
                    ),
                ),
                (
                    "related_user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="related_user_groups",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "root",
                    models.ForeignKey(
                        blank=True,
                        limit_choices_to=models.Q(("kind", "root")),
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="root_descendants",
                        to="django_woah.usergroup",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="membership",
            name="root_user_group",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("kind", "root")),
                on_delete=django.db.models.deletion.CASCADE,
                related_name="root_memberships",
                to="django_woah.usergroup",
            ),
        ),
        migrations.AddField(
            model_name="membership",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="memberships",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="membership",
            name="user_group",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="memberships",
                to="django_woah.usergroup",
            ),
        ),
        migrations.CreateModel(
            name="AssignedPerm",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid6.uuid7,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                    ),
                ),
                ("perm", models.CharField(max_length=128)),
                ("object_id", models.CharField(blank=True, max_length=40, null=True)),
                ("non_model_resource_id", models.TextField(blank=True, null=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="owned_assigned_perms",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user_group",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="group_assigned_perms",
                        to="django_woah.usergroup",
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="usergroup",
            constraint=models.UniqueConstraint(
                fields=("owner", "kind", "related_user"),
                name="unique_usergroup_owner_kind_related_user",
            ),
        ),
        migrations.AddConstraint(
            model_name="usergroup",
            constraint=models.UniqueConstraint(
                fields=("owner", "parent", "kind", "name"),
                name="unique_usergroup_localized_name",
            ),
        ),
        migrations.AddConstraint(
            model_name="membership",
            constraint=models.UniqueConstraint(
                fields=("user", "user_group"), name="unique_user_to_group_membership"
            ),
        ),
        migrations.AddIndex(
            model_name="assignedperm",
            index=models.Index(
                fields=["content_type", "object_id"],
                name="django_woah_content_b1c192_idx",
            ),
        ),
    ]



#  Copyright 2024 Pressinfra SRL
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from rest_framework.permissions import BasePermission


class IsAuthorized(BasePermission):
    """
    Allows access only to authenticated users.
    """

    def has_permission(self, request, view):
        if not (request.user and request.user.is_authenticated):
            return False

        if not hasattr(view, request.method.lower()):
            return view.http_method_not_allowed(request)

        if not hasattr(view, "action"):
            if request.method == "POST":
                # DRF has this "nice" way of testing if it should render a POST form in the default browsable API
                return view.is_authorized_for_unsaved_resource()

            return self.has_object_permission(
                request,
                view,
                obj=view.get_authorization_model_object(skip_authorization=True),
            )

        if getattr(view, "action", None):
            if view.action == "list" or (
                view.action == "metadata"
                and getattr(view, "authorize_options_as_get", True)
                and not view.detail
            ):
                # The queryset should be filtered accordingly for list
                # Metadata (OPTIONS) is handled the same when it's requested on a url where a GET
                # would produce a list action
                return True
            elif view.action == "create":
                return view.is_authorized_for_unsaved_resource()
            else:
                return self.has_object_permission(
                    request,
                    view,
                    obj=view.get_authorization_model_object(skip_authorization=True),
                )

        return True

    def has_object_permission(self, request, view, obj):
        # TODO: see what to do with "obj" parameter
        # return obj == view.get_authorization_model_object() breaks some cases

        if not hasattr(view, "action"):
            if request.method == "POST":
                # DRF has this "nice" way of testing if it should render a POST form in the default browsable API
                return view.is_authorized_for_unsaved_resource()

        elif view.action == "create":
            return view.is_authorized_for_unsaved_resource()

        obj = view.get_authorization_model_object()
        return obj is not None
